#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from . graphics import Graphics
from . frame import Frame, FrameProperty
from . path import Path


class Layout(Graphics):
    """
    Layout represents an experimental simple table-like layout manager tool.
    You can define custom layout by specifying rows and columns of fixed or
    relative size. Any pero.Graphics can then be inserted into specific cell.
    
    When the layout is drawn the cells are arranged according to available space
    and specific settings. The exact position of each cell is available via its
    'frame' property.
    
    Each cell graphics is drawn on temporarily modified canvas with specific
    drawing region (view) applied. This allows to use relative coordinates
    within the graphics. The exact position of each cell content is available
    via its 'content' property.
    
    Properties:
        
        x: int or float
            Specifies the x-coordinate of the top-left corner
        
        y: int or float
            Specifies the y-coordinate of the top-left corner
        
        width: int, float or UNDEF
            Specifies the full layout width. If set to UNDEF the full area of
            given canvas is used.
        
        height: int, float or UNDEF
            Specifies the full layout height. If set to UNDEF the full area of
            given canvas is used.
        
        padding: int, float or tuple
            Specifies the inner space of the layout as a single value or values
            for individual sides starting from top.
        
        spacing: int or float
            Specifies the space in-between individual cells.
        
        fill properties:
            Includes pero.FillProperties to specify the background fill.
        
        rows: (pero.Row,) (read-only)
            Gets rows definitions.
        
        cols: (pero.Column,) (read-only)
            Gets columns definitions.
        
        cells: (pero.Cell,) (read-only)
            Gets cells definitions.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    width = NumProperty(UNDEF)
    height = NumProperty(UNDEF)
    padding = QuadProperty(0)
    spacing = NumProperty(0)
    
    fill = Include(FillProperties, fill_color="w")
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Layout."""
        
        super().__init__(**overrides)
        
        self._rows = []
        self._cols = []
        self._cells = []
        self._grid = []
    
    
    @property
    def rows(self):
        """
        Gets rows definitions.
        
        Returns:
            (pero.Row,)
                Tuple of rows definitions.
        """
        
        return tuple(self._rows)
    
    
    @property
    def cols(self):
        """
        Gets columns definitions.
        
        Returns:
            (pero.Column,)
                Tuple of columns definitions.
        """
        
        return tuple(self._cols)
    
    
    @property
    def cells(self):
        """
        Gets cells definition.
        
        Returns:
            (pero.Cell,)
                Tuple of cells definitions.
        """
        
        return tuple(self._cells)
    
    
    def get_cell(self, row, col):
        """
        Gets the cell at specified layout grid position.
        
        Args:
            row: int
                Row index of requested cell.
            
            col: int
                Column index of requested cell.
        
        Returns:
            pero.Cell or None
                Corresponding cell or None.
        """
        
        return self._grid[row][col]
    
    
    def get_cell_below(self, x, y):
        """
        Gets the cell for which given coordinates fall into its bounding box.
        
        Args:
            x: int or float
                X-coordinate in logical units.
            
            y: int or float
                Y-coordinate in logical units.
        
        Returns:
            pero.Cell or None
                Corresponding cell or None.
        """
        
        for cell in self._cells:
            if cell.frame and cell.frame.contains(x, y):
                return cell
        
        return None
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw layout cells."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        
        # get size from canvas
        if width is UNDEF:
            width = canvas.viewport.width if canvas else 0
        if height is UNDEF:
            height = canvas.viewport.height if canvas else 0
        
        # arrange cells
        self.arrange(canvas, source, **overrides)
        
        # start drawing group
        canvas.group(tag, "layout")
        
        # set pen and brush
        canvas.line_width = 0
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw background
        canvas.draw_rect(x, y, width, height)
        
        # draw cells
        for cell in sorted(self._cells, key=lambda c: c.z_index):
            cell.draw(canvas)
        
        # end drawing group
        canvas.ungroup()
    
    
    def arrange(self, canvas=None, source=UNDEF, **overrides):
        """Calculates and sets cells frames."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        padding = self.get_property('padding', source, overrides)
        spacing = self.get_property('spacing', source, overrides)
        
        # get size from canvas
        if width is UNDEF:
            width = canvas.viewport.width if canvas else 0
        if height is UNDEF:
            height = canvas.viewport.height if canvas else 0
        
        # apply padding and spacing
        padding = padding or (0, 0, 0, 0)
        x += padding[3]
        y += padding[0]
        width -= spacing * (len(self._cols) - 1) + padding[1] + padding[3]
        height -= spacing * (len(self._rows) - 1) + padding[0] + padding[2]
        
        # arrange rows and cols
        heights = self._arrange_rows(height, spacing)
        widths = self._arrange_cols(width, spacing)
        
        # arrange cells
        for cell in self._cells:
            
            r, rs = cell.row, cell.row_span
            c, cs = cell.col, cell.col_span
            
            cell.frame = Frame(
                x = x + sum(widths[0:c]) + spacing * c,
                y = y + sum(heights[0:r]) + spacing * r,
                width = sum(widths[c:c+cs]) + spacing * (cs-1),
                height = sum(heights[r:r+rs]) + spacing * (rs-1))
            
            cell.arrange(canvas)
    
    
    def add(self, graphics, row, col, **overrides):
        """
        Adds graphics to specified layout cell. Additional rows and columns are
        added if necessary with relative size of 1.
        
        Args:
            graphics: pero.Graphics
                Graphics to be added.
            
            row: int
                Index of the row into which the graphics should be added.
            
            col: int
                Index of the column into which the graphics should be added.
            
            row_span: int
                Number of rows the graphics should span.
            
            col_span: int
                Number of columns the graphics should span.
            
            width: int, float or UNDEF
                Minimum width of the content area. If not specified, the content
                may use whole available width.
            
            height: int, float or UNDEF
                Minimum height of the content area. If not specified, the
                content may use whole available height.
            
            padding: int, float or tuple
                Inner empty space of the cell as a single value or values for
                individual sides starting from top.
            
            h_expand: bool
                If set to True the cell content can use full available width
                even if the 'width' property is set.
            
            v_expand: bool
                If set to True the cell content can use full available height
                even if the 'height' property is set.
            
            h_align: str
                Horizontal alignment of the cell content as any item from the
                pero.POSITION_LRC enum. Note that for this to have any effect
                the 'width' property must be specified.
            
            v_align: str
                Vertical alignment of the cell content as any item from the
                pero.POSITION_TBC enum. Note that for this to have any effect
                the 'height' property must be specified.
            
            line properties:
                All the pero.LineProperties to specify the cell background
                outline.
            
            fill properties:
                All the pero.FillProperties to specify the cell background
                fill.
        """
        
        # check type
        if not isinstance(graphics, Graphics):
            message = "Graphics must be of type pero.Graphics! -> %s" % type(graphics)
            raise TypeError(message)
        
        # init cell
        cell = Cell(
            graphics = graphics,
            row = row,
            col = col,
            **overrides)
        
        # get used rows and columns
        rows = cell.rows
        cols = cell.cols
        
        # add rows and columns if necessary
        while len(self._rows) <= max(rows):
            self.add_row(1, True)
        
        while len(self._cols) <= max(cols):
            self.add_col(1, True)
        
        # check if cells are empty
        for r in rows:
            for c in cols:
                if self._grid[r][c] is not None:
                    message = "Specified cell (%s,%s) is already used!" % (r, c)
                    raise ValueError(message)
        
        # register cell
        self._cells.append(cell)
        
        # add cell to grid
        for r in rows:
            for c in cols:
                self._grid[r][c] = cell
    
    
    def add_row(self, height=1, relative=True):
        """
        Adds additional row at the end of layout with specified absolute or
        relative height.
        
        Args:
            height: int or float
                Absolute or relative height of the row.
            
            relative:
                If set to True, specified height is considered as relative
                portion of total available space after filling all fixed rows.
        """
        
        # init row
        row = Row(
            height = height,
            relative = relative)
        
        # add row
        self._rows.append(row)
        
        # update grid
        self._grid.append([None for c in range(len(self._cols))])
    
    
    def add_col(self, width=1, relative=True):
        """
        Adds additional column at the end of layout with specified absolute or
        relative width.
        
        Args:
            width: int or float
                Absolute or relative width of the column.
            
            relative:
                If set to True, specified width is considered as relative
                portion of total available space after filling all fixed
                columns.
        """
        
        # init column
        col = Column(
            width = width,
            relative = relative)
        
        # add column
        self._cols.append(col)
        
        # update grid
        for row in self._grid:
            row.append(None)
    
    
    def _arrange_rows(self, available, spacing):
        """Calculates final size of each row."""
        
        # init buffers
        heights = [0]*len(self._rows)
        minima = [0]*len(self._rows)
        dirty = False
        
        # get sums
        fix_sum = sum(r.height for r in self._rows if not r.relative)
        rel_sum = sum(r.height for r in self._rows if r.relative)
        
        # calc naive heights and minima
        for i, row in enumerate(self._rows):
            
            # set fixed row
            if not row.relative:
                heights[i] = row.height
                minima[i] = row.height
            
            # set relative row
            else:
                heights[i] = max(0, (available - fix_sum) * row.height / rel_sum)
                cells = [c for c in self._grid[i] if c and c.height and c.row_span == 1]
                minima[i] = max(c.height for c in cells) if cells else 0
            
            # check minimum height
            dirty = dirty or heights[i] < minima[i]
        
        # force minima
        while dirty:
            dirty = False
            
            for i, row in enumerate(self._rows):
                
                if heights[i] < minima[i]:
                    heights[i] = minima[i]
                    fix_sum += minima[i]
                    rel_sum -= row.height
                    dirty = True
                    break
                
                if heights[i] > minima[i] and row.relative:
                    heights[i] = max(0, (available - fix_sum) * row.height / rel_sum)
        
        return heights
    
    
    def _arrange_cols(self, available, spacing):
        """Calculates final size of each column."""
        
        # init buffers
        widths = [0]*len(self._cols)
        minima = [0]*len(self._cols)
        dirty = False
        
        # get sums
        fix_sum = sum(c.width for c in self._cols if not c.relative)
        rel_sum = sum(c.width for c in self._cols if c.relative)
        
        # calc naive widths and minima
        for i, col in enumerate(self._cols):
            
            # set fixed col
            if not col.relative:
                widths[i] = col.width
                minima[i] = col.width
            
            # set relative col
            else:
                widths[i] = max(0, (available - fix_sum) * col.width / rel_sum)
                cells = [self._grid[r][i] for r in range(len(self._rows))]
                cells = [c for c in cells if c and c.width and c.col_span == 1]
                minima[i] = max(c.width for c in cells) if cells else 0
            
            # check minimum height
            dirty = dirty or widths[i] < minima[i]
        
        # force minima
        while dirty:
            dirty = False
            
            for i, col in enumerate(self._cols):
                
                if widths[i] < minima[i]:
                    widths[i] = minima[i]
                    fix_sum += minima[i]
                    rel_sum -= col.width
                    dirty = True
                    break
                
                if widths[i] > minima[i] and col.relative:
                    widths[i] = max(0, (available - fix_sum) * col.width / rel_sum)
        
        return widths


class Row(PropertySet):
    """
    Layout row definition. This is typically created automatically by layout
    manager but the properties can be later changed if needed.
    
    Properties:
        
        height: int or float
            Specifies absolute or relative height of the row.
        
        relative: bool
            Specifies whether the row height is considered as relative portion
            of total available space after filling all fixed rows.
    """
    
    height = NumProperty(UNDEF, dynamic=False)
    relative = BoolProperty(False, dynamic=False)


class Column(PropertySet):
    """
    Layout column definition. This is typically created automatically by layout
    manager but the properties can be later changed if needed.
    
    Properties:
        
        width: int or float
            Specifies absolute or relative width of the column.
        
        relative: bool
            Specifies whether the column width is considered as relative portion
            of total available space after filling all fixed columns.
    """
    
    width = NumProperty(UNDEF, dynamic=False)
    relative = BoolProperty(False, dynamic=False)


class Cell(Graphics):
    """
    Layout cell definition. This is typically created automatically by layout
    manager but some of the properties can be later changed if needed.
    
    Properties:
        
        graphics: pero.Graphics
            Specifies the graphics to drawn.
        
        row: int
            Specifies the index of the row in which the cell sits. This value
            cannot be changed once set.
        
        col: int
            Specifies the index of the column in which the cell sits. This value
            cannot be changed once set.
        
        row_span: int
            Specifies the number of rows the cell spans. This value cannot be
            changed once set.
        
        col_span: int
            Specifies the number of columns the cell spans. This value cannot be
            changed once set.
        
        clip: bool
            Specifies whether the content overflow should be clipped.
        
        width: int, float or UNDEF
            Specifies the minimum width of the content area. If not specified,
            the content may use whole available width.
        
        height: int, float or UNDEF
            Specifies the minimum height of the content area. If not specified,
            the content may use whole available height.
        
        padding: int, float or tuple
            Specifies the inner empty space of the cell as a single value or
            values for individual sides starting from top.
        
        h_expand: bool
            Specifies whether the cell content can use full available width even
            if the 'width' property is set.
        
        v_expand: bool
            Specifies whether the cell content can use full available height
            even if the 'height' property is set.
        
        h_align: str
            Specifies the horizontal alignment of the cell content as any item
            from the pero.POSITION_LRC enum. Note the for this to have any
            effect the 'width' property must be specified and 'h_expand' must be
            disabled.
        
        v_align: str
            Specifies vertical alignment of the cell content as any item from
            the pero.POSITION_TBC enum. Note the for this to have any effect
            the 'height' property must be specified and 'v_expand' must be
            disabled.
        
        line properties:
            Includes pero.LineProperties to specify the cell background
            outline.
        
        fill properties:
            Includes pero.FillProperties to specify the cell background fill.
        
        frame: pero.Frame
            Specifies the cell frame. This value is typically calculated and set
            automatically by the layout manager.
        
        content: pero.Frame
            Specifies the cell content frame. This value is typically calculated
            and set automatically by the layout manager.
    """
    
    graphics = Property(UNDEF, types=(Graphics,), dynamic=False)
    
    row = IntProperty(UNDEF, dynamic=False)
    col = IntProperty(UNDEF, dynamic=False)
    row_span = IntProperty(1, dynamic=False)
    col_span = IntProperty(1, dynamic=False)
    
    width = NumProperty(UNDEF, dynamic=False)
    height = NumProperty(UNDEF, dynamic=False)
    padding = QuadProperty(0, dynamic=False)
    clip = BoolProperty(True)
    
    h_expand = BoolProperty(True)
    v_expand = BoolProperty(True)
    
    h_align = EnumProperty(POS_CENTER, enum=POSITION_LRC, dynamic=False)
    v_align = EnumProperty(POS_CENTER, enum=POSITION_TBC, dynamic=False)
    
    outline = Include(LineProperties, line_color=None)
    fill = Include(FillProperties, fill_color=None)
    
    frame = FrameProperty(UNDEF, dynamic=False)
    content = FrameProperty(UNDEF, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of layout Cell."""
        
        super().__init__(**overrides)
        
        self._content_origin = (0, 0)
        
        # lock properties
        self.lock_property('row')
        self.lock_property('col')
        self.lock_property('row_span')
        self.lock_property('col_span')
    
    
    @property
    def rows(self):
        """
        Gets indices of all the rows this cell spans.
        
        Returns:
            (int,)
                Tuple of rows indices.
        """
        
        return tuple(self.row+s for s in range(self.row_span))
    
    
    @property
    def cols(self):
        """
        Gets indices of all the columns this cell spans.
        
        Returns:
            (int,)
                Tuple of columns indices.
        """
        
        return tuple(self.col+s for s in range(self.col_span))
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw cell graphics."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        graphics = self.get_property('graphics', source, overrides)
        frame = self.get_property('frame', source, overrides)
        content = self.get_property('content', source, overrides)
        clip = self.get_property('clip', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # start drawing group
        canvas.group(tag, "layout_cell")
        
        # draw background
        canvas.draw_rect(*frame.rect)
        
        # set clipping
        if clip:
            canvas.clip(Path().rect(*content.rect))
        
        # draw graphics
        if graphics:
            with canvas.view(*content.rect, relative=True):
                graphics.draw(canvas)
        
        # revert clipping
        if clip:
            canvas.unclip()
        
        # end drawing group
        canvas.ungroup()
    
    
    def arrange(self, canvas=None, source=UNDEF, **overrides):
        """Calculates and sets content frame."""
        
        # get properties
        frame = self.get_property('frame', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        padding = self.get_property('padding', source, overrides)
        h_expand = self.get_property('h_expand', source, overrides)
        v_expand = self.get_property('v_expand', source, overrides)
        h_align = self.get_property('h_align', source, overrides)
        v_align = self.get_property('v_align', source, overrides)
        clip = self.get_property('clip', source, overrides)
        
        # get coords
        padding = padding or (0, 0, 0, 0)
        width = width or 0
        height = height or 0
        
        if not width or h_expand:
            width = max(width, frame.width - padding[1] - padding[3])
        if not height or v_expand:
            height = max(height, frame.height - padding[0] - padding[2])
        
        if h_align == POS_CENTER:
            x = frame.cx + padding[3] - padding[1] - .5*width
        elif h_align == POS_RIGHT:
            x = frame.x2 - width - padding[1]
        else:
            x = frame.x + padding[3]
        
        if v_align == POS_CENTER:
            y = frame.cy + padding[0] - padding[2] - .5*height
        elif v_align == POS_BOTTOM:
            y = frame.y2 - height - padding[2]
        else:
            y = frame.y + padding[0]
        
        # apply clipping
        if clip:
            x = max(x, frame.x + padding[3])
            y = max(y, frame.y + padding[0])
            width = min(width, frame.width - padding[1] - padding[3])
            height = min(height, frame.height - padding[0] - padding[2])
        
        # check flipped content
        width = max(0, width)
        height = max(0, height)
        
        # set content frame
        self.content = Frame(x, y, width, height)
    
    
    def to_content(self, x, y):
        """
        Recalculates given position from the layout coordinate system into the
        cell content system.
        
        Args:
            x: float
                X-coordinate to convert.
            
            y: float
                Y-coordinate to convert.
        
        Returns:
            (float, float)
                X and y coordinates within cell content coordinate system.
        """
        
        return x - self.content.x, y - self.content.y
    
    
    def to_layout(self, x, y):
        """
        Recalculates given position from the cell content coordinate system into
        the layout system.
        
        Args:
            x: float
                X-coordinate to convert.
            
            y: float
                Y-coordinate to convert.
        
        Returns:
            (float, float)
                X and y coordinates within layout coordinate system.
        """
        
        return x + self.content.x, y + self.content.y



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy


class Matrix(object):
    """
    Represents 2D transformation matrix used to apply requested transformations
    to given coordinates, mostly to transform a path.
    
    Most of the methods returns self so that they can be chained.
    """
    
    
    def __init__(self):
        """Initializes a new instance of Matrix."""
        
        self._matrix = numpy.identity(3)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return str(self._matrix)
    
    
    def clear(self):
        """
        Clears all applied transformations and resets current matrix to identity
        matrix.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        self._matrix = numpy.identity(3)
        
        return self
    
    
    def translate(self, x_shift=None, y_shift=None):
        """
        Applies horizontal and vertical shift to current matrix.
        
        Args:
            x_shift: int or float
                Horizontal shift.
            
            y_shift: int or float
                Vertical shift.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        if not x_shift and not y_shift:
            return self
        
        elif y_shift is None:
            y_shift = 0
        
        elif x_shift is None:
            x_shift = 0
        
        trans = numpy.array((
            [1.0, 0.0, x_shift],
            [0.0, 1.0, y_shift],
            [0.0, 0.0, 1.0]))
        
        self._matrix = numpy.dot(trans, self._matrix)
        
        return self
    
    
    def rotate(self, angle, x=0, y=0, clockwise=True):
        """
        Applies rotation around specified origin to current matrix.
        
        Args:
            angle: float
                Angle to be applied in radians.
            
            x: int float
                X-coordinate of the rotation origin.
            
            y: int float
                Y-coordinate of the rotation origin.
            
            clockwise: bool
                Specifies the direction of rotation. If set to True, the rotation
                will be applied in the clockwise direction.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        clockwise = 1 if clockwise else -1
        
        a = numpy.cos(angle) * clockwise
        b = numpy.sin(angle) * clockwise
        
        trans = numpy.array((
            [a, -b, 0.0],
            [b, a, 0.0],
            [0.0, 0.0, 1.0]))
        
        if x != 0 or y != 0:
            self.translate(-x, -y)
        
        self._matrix = numpy.dot(trans, self._matrix)
        
        if x != 0 or y != 0:
            self.translate(x, y)
        
        return self
    
    
    def scale(self, x_scale, y_scale, x=0, y=0):
        """
        Applies horizontal and vertical scaling from specified origin to current
        matrix.
        
        Args:
            x_scale: int or float
                Horizontal axis scale.
            
            y_scale: int or float
                Vertical axis scale.
            
            x: int float
                X-coordinate of the scaling origin.
            
            y: int float
                Y-coordinate of the scaling origin.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        trans = numpy.array((
            [x_scale, 0.0, 0.0],
            [0.0, y_scale, 0.0],
            [0.0, 0.0, 1.0]))
        
        if x != 0 or y != 0:
            self.translate(-x, -y)
        
        self._matrix = numpy.dot(trans, self._matrix)
        
        if x != 0 or y != 0:
            self.translate(x, y)
        
        return self
    
    
    def skew(self, x_shear, y_shear, x=0, y=0, clockwise=True):
        """
        Applies skewing and shearing around specified origin to current matrix.
        
        Args:
            x_shear: float
                Angle to be applied along horizontal axis in radians.
            
            y_shear: float
                Angle to be applied along vertical axis in radians.
            
            x: int float
                X-coordinate of the rotation origin.
            
            y: int float
                Y-coordinate of the rotation origin.
            
            clockwise: bool
                Specifies the direction of skew. If set to True, the skew
                will be applied in the clockwise direction.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        clockwise = 1 if clockwise else -1
        
        a = numpy.tan(x_shear) * clockwise
        b = numpy.tan(y_shear) * clockwise
        
        trans = numpy.array((
            [1.0, a, 0.0],
            [b, 1.0, 0.0],
            [0.0, 0.0, 1.0]))
        
        if x != 0 or y != 0:
            self.translate(-x, -y)
        
        self._matrix = numpy.dot(trans, self._matrix)
        
        if x != 0 or y != 0:
            self.translate(x, y)
        
        return self
    
    
    def ray(self, length, angle):
        """
        Applies shift along the direction of given angle to current matrix.
        
        Args:
            length: float
                Shift along the angle.
            
            angle: float
                Angle of the shift in radians.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        self.rotate(-angle)
        self.translate(length)
        self.rotate(angle)
        
        return self
    
    
    def hflip(self, x=0):
        """
        Applies horizontal flipping along specified x-coordinate to current
        matrix.
        
        Args:
            x: int or float
                Flipping x-coordinate.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        self.scale(-1, 1, x, 0)
        
        return self
    
    
    def vflip(self, y=0):
        """
        Applies vertical flipping along specified y-coordinate to current
        matrix.
        
        Args:
            y: int or float
                Flipping y-coordinate.
        
        Returns:
            pero.Matrix
                Returns self so that the commands can be chained.
        """
        
        self.scale(1, -1, 0, y)
        
        return self
    
    
    def transform(self, x, y):
        """
        Transforms given point by current matrix.
        
        Args:
            x: int or float
                X-coordinate of the point to be transformed.
            
            y: int or float
                Y-coordinate of the point to be transformed.
        
        Returns:
            (int, int) or (float, float)
                Transformed point.
        """
        
        return numpy.array((x, y, 1)).dot(self._matrix.T)[0:2]



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from .. drawing import Path, FrameProperty
from . glyph import Glyph
from . markers import Marker


class Band(Glyph):
    """
    Band glyph provides basic functionality to draw x-sorted data points as a
    filled band of varying thickness.
    
    To plot data as a line the 'show_line' property must be set to True.
    Similarly, to display individual points, the 'show_points' property must be
    set to True. If set to UNDEF, and line is enabled, the points are
    automatically visible if they are separated enough, as defined by the
    'spacing' property. The band area can be displayed if 'show_area' property
    is set to True.
    
    When individual points are drawn, original 'data' item is provided as the
    'source' together with relevant overrides to the marker glyph so any
    property of the marker can be dynamic.
    
    If the 'clip' property is defined, drawn data will be clipped to show the
    specified region only. Note that this is applied to points only and not to
    the line and area fill.
    
    Properties:
        
        show_line: bool or callable
            Specifies whether the profile lines should be displayed.
        
        show_points: bool, callable or UNDEF
            Specifies whether individual points should be displayed. If set to
            UNDEF, the points are displayed automatically as long as they are
            distance is big enough.
        
        show_area: bool or callable
            Specifies whether the area between the profile lines should be
            displayed.
        
        data: tuple, callable, None or UNDEF
            Specifies the sequence of raw data to be provided as the source for
            drawing individual points.
        
        x: tuple or callable
            Specifies the x-coordinates of the profile lines.
        
        y1: tuple or callable
            Specifies the top y-coordinates of the profile line.
        
        y2: tuple or callable
            Specifies the bottom y-coordinates of the profile line.
        
        spacing: int, float, callable
            Specifies the minimum x-distance between points to enable automatic
            points display.
        
        clip: pero.Frame, callable, None or UNDEF
            Specifies the clipping frame to skip drawing of points outside the
            frame.
        
        marker: pero.MARKER, pero.Path, pero.Marker, callable, None or UNDEF
            Specifies the marker to draw actual data points with. The value
            can be specified by any item from the pero.MARKER enum, as symbol
            pero.Path or as pero.Marker directly.
        
        marker_size: int, float or callable
            Specifies the marker size.
        
        marker_line properties:
            Includes pero.LineProperties to specify the marker line.
        
        marker_fill properties:
            Includes pero.FillProperties to specify the marker fill.
        
        line properties:
            Includes pero.LineProperties to specify the profile line.
        
        fill properties:
            Includes pero.FillProperties to specify the area fill.
    """
    
    show_line = BoolProperty(True)
    show_points = BoolProperty(UNDEF)
    show_area = BoolProperty(False)
    
    data = SequenceProperty(UNDEF, nullable=True)
    x = SequenceProperty(UNDEF, intypes=(int, float))
    y1 = SequenceProperty(UNDEF, intypes=(int, float))
    y2 = SequenceProperty(UNDEF, intypes=(int, float))
    
    spacing = IntProperty(10)
    clip = FrameProperty(UNDEF)

    marker = Property(MARKER_CIRCLE, types=(str, Path, Marker), nullable=True)
    marker_size = NumProperty(4)
    marker_line = Include(LineProperties, prefix='marker_', line_color=UNDEF)
    marker_fill = Include(FillProperties, prefix='marker_', fill_color=UNDEF)
    
    line = Include(LineProperties, line_color=UNDEF)
    fill = Include(FillProperties, fill_color=UNDEF)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        show_line = self.get_property('show_line', source, overrides)
        show_points = self.get_property('show_points', source, overrides)
        show_area = self.get_property('show_area', source, overrides)
        spacing = self.get_property('spacing', source, overrides)
        x_coords = self.get_property('x', source, overrides)
        y1_coords = self.get_property('y1', source, overrides)
        y2_coords = self.get_property('y2', source, overrides)
        
        # enable/disable points display
        if show_points is UNDEF:
            diff = numpy.min(numpy.diff(x_coords)) if len(x_coords) > 1 else 0
            show_points = diff > spacing
        
        # start drawing group
        canvas.group(tag, "band")
        
        # draw area
        if show_area:
            self._draw_area(canvas, source, overrides, x_coords, y1_coords, y2_coords)
        
        # draw lines
        if show_line:
            self._draw_lines(canvas, source, overrides, x_coords, y1_coords, y2_coords)
        
        # draw points
        if show_points:
            self._draw_points(canvas, source, overrides, x_coords, y1_coords, y2_coords)
        
        # end drawing group
        canvas.ungroup()
    
    
    def _draw_area(self, canvas, source, overrides, x_coords, y1_coords, y2_coords):
        """Draws band area."""
        
        # set pen and brush
        canvas.line_color = None
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # concatenate points
        x_coords = numpy.concatenate((x_coords, numpy.flip(x_coords, 0)), 0)
        y_coords = numpy.concatenate((y1_coords, numpy.flip(y2_coords, 0)), 0)
        
        # init points
        points = numpy.stack((x_coords, y_coords), axis=1)
        
        # draw polygon
        canvas.draw_polygon(points)
    
    
    def _draw_lines(self, canvas, source, overrides, x_coords, y1_coords, y2_coords):
        """Draws band lines."""
        
        # set pen and brush
        canvas.fill_color = None
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw lines
        points = numpy.stack((x_coords, y1_coords), axis=1)
        canvas.draw_lines(points)
        
        points = numpy.stack((x_coords, y2_coords), axis=1)
        canvas.draw_lines(points)
    
    
    def _draw_points(self, canvas, source, overrides, x_coords, y1_coords, y2_coords):
        """Draws individual points."""
        
        # get properties
        clip = self.get_property('clip', source, overrides)
        data = self.get_property('data', source, overrides)
        
        # get marker overrides
        marker_overrides = self.get_child_overrides('marker', overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # check raw data
        has_data = data is not UNDEF and data is not None and len(data) == len(x_coords)
        
        # draw points
        for i in range(len(x_coords)):
            
            # get point data
            point_data = data[i] if has_data else None
            
            # get marker
            marker = self.get_property('marker', point_data, overrides)
            if not marker:
                continue
            
            # init glyph
            if not isinstance(marker, Marker):
                marker = Marker.create(marker)
                marker.set_properties_from(self, src_prefix='marker_', overrides=overrides, native=True)
            
            # get coords
            x = x_coords[i]
            y1 = y1_coords[i]
            y2 = y2_coords[i]
            radius = 0.5 * marker.get_property('size', point_data, marker_overrides)
            
            # apply x-clipping
            if clip and (clip.x1 > x+radius or clip.x2 < x-radius):
                continue
            
            # draw y1
            if not clip or (clip.y1 < y1+radius and clip.y2 > y1-radius):
                marker.draw(canvas, point_data, x=x, y=y1, **marker_overrides)
            
            # draw y2
            if not clip or (clip.y1 < y2+radius and clip.y2 > y2-radius):
                marker.draw(canvas, point_data, x=x, y=y2, **marker_overrides)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . control import Control, ToolControl
from . sizer import Sizer
from . tool import Tool
from . view import View
from . export import show, export, debug
from . json import Image

# import main backends
from . import json
from . import svg

# try import backends
try: from . import cairo
except ImportError: pass

try: from . import mupdf
except ImportError: pass

try: from . import qt
except ImportError: pass

try: from . import wx
except ImportError: pass

try: from . import pythonista
except ImportError: pass



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy


# comparison with tolerance

def equals(v1, v2, epsilon):
    """
    Returns True if difference between given values is less then tolerance.
    
    Args:
        v1: float
            Value one.
        
        v2: float
            Value two
        
        epsilon: float
            Max allowed difference.
    
    Returns:
        bool
            True if values equal, False otherwise.
    """
    
    return abs(v1-v2) <= epsilon


def between(v, min_v, max_v, epsilon):
    """
    Returns True if given value is equal or between minimum and maximum by
    specified tolerance.
    
    Args:
        v: float
            Value to check.
        
        min_v: float
            Minimum value.
        
        max_v: float
            Maximum value.
        
        epsilon: float
            Max allowed difference.
    
    Returns:
        bool
            True if value is equal or between, False otherwise.
    """
    
    return (min_v <= v <= max_v) or equals(v, min_v, epsilon) or equals(v, max_v, epsilon)


# angle calculations

def rads(angle):
    """
    Converts given angle from degrees to radians.
    
    Args:
        angle: float
            Angle in degrees.
    
    Returns:
        float
            Angle in radians.
    """
    
    return numpy.deg2rad(angle)


def degs(angle):
    """
    Converts given angle from radians to degrees.
    
    Args:
        angle: float
            Angle in radians.
    
    Returns:
        float
            Angle in degrees.
    """
    
    return numpy.rad2deg(angle)


def angle(p1, p2, p3):
    """
    Calculates angle between two lines.
    
    Args:
        p1: (float, float)
            First point as (x, y) coordinates.
        
        p2: (float, float)
            Origin point as (x, y) coordinates.
        
        p3: (float, float)
            Second point as (x, y) coordinates.
    
    Returns:
        float
            Angle in radians.
    """
    
    dx1 = p1[0] - p2[0]
    dy1 = p1[1] - p2[1]
    dx2 = p3[0] - p2[0]
    dy2 = p3[1] - p2[1]
    
    cross = dx1*dy2 - dy1*dx2
    dot = dx1*dx2 + dy1*dy2
    
    return numpy.arctan2(cross, dot)


def inclination(p1, p2):
    """
    Calculates inclination angle the line has with x-axis.
    
    Args:
        p1: (float, float)
            Origin point as (x, y) coordinates.
        
        p2: (float, float)
            Second point as (x, y) coordinates.
    
    Returns:
        float
            Angle in radians.
    """
    
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    
    return numpy.arctan2(dy, dx)


def bisector(p1, p2, p3):
    """
    Calculates bisector angle between two lines.
    
    Args:
        p1: (float, float)
            First point as (x, y) coordinates.
        
        p2: (float, float)
            Origin point as (x, y) coordinates.
        
        p3: (float, float)
            Second point as (x, y) coordinates.
    
    Returns:
        float
            Angle in radians.
    """
    
    a1 = inclination(p2, p1)
    a2 = inclination(p2, p3)
    
    return 0.5 * (a1 + a2)


def normal_angle(angle):
    """
    Normalizes given angle to be in interval of -2pi to 2pi.
    
    Args:
        angle: float
            Angle in radians.
    
    Returns:
        float
            Normalized angle in radians.
    """
    
    return angle % (2*numpy.pi)


def angle_difference(start_angle, end_angle, clockwise):
    """
    Calculates difference between two given angles.
    
    Args:
        start_angle: float
            Start angle in radians.
        
        end_angle: float
            End angle in radians.
        
        clockwise: bool
            Specifies the direction of measurement.
    
    Returns:
        float
            Angle difference in radians.
    """
    
    start = normal_angle(start_angle)
    end = normal_angle(end_angle)
    
    diff = end-start if clockwise else start-end
    if diff < 0:
        diff = 2 * numpy.pi - abs(diff)
    
    if not clockwise:
        diff *= -1
    
    return diff


# points calculations

def distance(p1, p2):
    """
    Calculates Euclidean distance between two points.
    
    Args:
        p1: (float, float)
            Point 1 as (x, y) coordinates.
        
        p2: (float,float)
            Point 2 as (x, y) coordinates.
    
    Returns:
        float
            Distance between the points.
    """
    
    dx = p1[0] - p2[0]
    dy = p1[1] - p2[1]
    sq = dx*dx + dy*dy
    
    return numpy.sqrt(sq) if sq > 0 else 0


def rotate(p, angle, center=(0, 0)):
    """
    Rotates given point around specified center.
    
    Args:
        p: (float, float)
            Point to rotate.
        
        angle: float
            Angle in radians.
        
        center: (float, float)
            Center of rotation.
    """
    
    dx = p[0]-center[0]
    dy = p[1]-center[1]
    
    sin = numpy.sin(angle)
    cos = numpy.cos(angle)
    
    x = center[0] + dx * cos - dy * sin
    y = center[1] + dx * sin + dy * cos
    
    return x, y


def ray(c, angle, distance):
    """
    Calculates point coordinates with distance and angle from origin.
    
    Args:
        c: (float, float)
            Coordinates of the origin.
        
        angle: float
            Angle in radians.
        
        distance: float
            Distance from origin.
    
    Returns:
            (float, float)
                Coordinates of calculated point.
    """
    
    x = c[0] + distance * numpy.cos(angle)
    y = c[1] + distance * numpy.sin(angle)
    
    return x, y


def polygon_centroid(*points):
    """
    Calculates center point of given polygon.
    
    Args:
        points: ((float, float),)
            Collection of points as (x, y) coordinates.
    
    Returns:
        (float, float)
            Center point as (x, y) coordinates.
    """
    
    x = sum(p[0] for p in points) / len(points)
    y = sum(p[1] for p in points) / len(points)
    
    return x, y


def triangle_incircle(p1, p2, p3):
    """
    Calculates position and size of the biggest circle inscribed into
    specified triangle.
    
    Args:
        p1: (float, float)
            Point 1 as (x, y) coordinates.
        
        p2: (float, float)
            Point 2 as (x, y) coordinates.
        
        p3: (float, float)
            Point 3 as (x, y) coordinates.
    
    Returns:
        (float, float)
            Coordinates of the circle center.
        
        float
            Circle radius.
    """
    
    # calc sides
    a = distance(p1, p2)
    b = distance(p2, p3)
    c = distance(p3, p1)
    
    # calc radius
    p = 0.5 * sum((a, b, c))
    area = numpy.sqrt(p * (p - a) * (p - b) * (p - c))
    radius = area / p
    
    # calc center
    bis1 = bisector(p2, p1, p3)
    bis2 = bisector(p1, p2, p3)
    c = intersect_rays(p1, bis1, p2, bis2)
    
    return c, radius


# position calculations

def is_inline(*points):
    """
    Checks whether all given points are on a single line. The points order
    is not important.
    
    Args:
        *points: ((float, float),)
            Collection of points to check as (x,y) coordinates.
    
    Returns:
        bool
            Returns True if all points are on a single line, False otherwise.
    """
    
    if len(points) < 3:
        return True
    
    x1, y1 = points[0]
    x2, y2 = points[1]
    
    dx1 = x1 - x2
    dy1 = y1 - y2
    
    for x2, y2 in points[2:]:
        dx2 = x1 - x2
        dy2 = y1 - y2
        if (dy1*dx2) != (dx1*dy2):
            return False
    
    return True


def is_point_in_circle(p, center, radius):
    """
    Checks whether given point is within specified circle.
    
    Args:
        p: (float, float)
            Coordinates of the point to test.
        
        center: (float, float)
            Coordinates of the circle center.
        
        radius: float
            Radius of the circle.
    
    Returns:
        bool
            Returns True if the point is inside the circle, False otherwise.
    """
    
    return distance(center, p) < radius


def is_point_in_triangle(p, p1, p2, p3):
    """
    Checks whether given point is within specified triangle.
    
    Args:
        p: (float, float)
            Coordinates of the point to test.
        
        p1: (float, float)
            Coordinates of the triangle point.
        
        p2: (float, float)
            Coordinates of the triangle point.
        
        p3: (float, float)
            Coordinates of the triangle point.
    
    Returns:
        bool
            Returns True if the point is inside the circle, False otherwise.
    """
    
    xp, yp = p
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    
    c1 = (x2-x1)*(yp-y1)-(y2-y1)*(xp-x1)
    c2 = (x3-x2)*(yp-y2)-(y3-y2)*(xp-x2)
    c3 = (x1-x3)*(yp-y3)-(y1-y3)*(xp-x3)
    
    return (c1 < 0 and c2 < 0 and c3 < 0) or (c1 > 0 and c2 > 0 and c3 > 0)


def is_point_in_polygon(p, *polygon):
    """
    Checks whether given point is within specified polygon. This technique
    is bases on the sum of all angles between given point and polygon vertices,
    which must be 2pi if point is inside. This assumption only works for non-
    overlapping polygons.
    
    Args:
        p: (float, float)
            Coordinates of the point to test.
        
        *polygon: ((float, float),)
            Collection of points defining the polygon as (x,y) coordinates.
    
    Returns:
        bool
            Returns True if the point lies inside the circle, False otherwise.
    """
    
    total_angle = 0
    polygon = list(polygon) + [polygon[0]]
    p1 = polygon[0]
    
    for p2 in polygon[1:]:
        total_angle += abs(angle(p1, p, p2)) % numpy.pi
        p1 = p2
    
    return equals(total_angle, 2*numpy.pi, 1e-6)


def is_circle_in_circle(c1, r1, c2, r2):
    """
    Checks whether the first circle is inside the second circle..
    
    Args:
        c1: (float, float)
            Coordinates of the center of circle A.
        
        r1: float
            Radius of the circle A.
        
        c2: (float, float)
            Coordinates of the center of circle B.
        
        r2: float
            Radius of the circle b.
    
    Returns:
        ((float, float), (float, float)) or None
            Coordinates of the two intersection points.
    """
    
    return distance(c1, c2) <= r2 - r1


# intersections calculations

def intersect_circles(c1, r1, c2, r2):
    """
    Calculates intersection points between two circles.
    
    Args:
        c1: (float, float)
            Coordinates of the center of circle A.
        
        r1: float
            Radius of the circle A.
        
        c2: (float, float)
            Coordinates of the center of circle B.
        
        r2: float
            Radius of the circle b.
    
    Returns:
        ((float, float), (float, float)) or None
            Coordinates of the two intersection points.
    """
    
    # calc distance
    dx = c2[0] - c1[0]
    dy = c2[1] - c1[1]
    dist = numpy.sqrt(dx*dx + dy*dy)
    
    # non intersecting
    if dist >= r1 + r2:
        return ()
    
    # one inside another
    if dist <= abs(r1-r2):
        return ()
    
    # same circles
    if dist == 0 and r1 == r2:
        return ()
    
    # calc intersections
    a = (r1**2 - r2**2 + dist**2) / (2*dist)
    n = r1**2 - a**2
    h = numpy.sqrt(n if n > 0 else 0)
    
    x = c1[0] + a*dx / dist
    y = c1[1] + a*dy / dist
    
    x1 = x + h*dy / dist
    y1 = y - h*dx / dist
    x2 = x - h*dy / dist
    y2 = y + h*dx / dist
    
    return (x1, y1), (x2, y2)


def intersect_lines(p1, p2, p3, p4):
    """
    Calculates intersection point between two lines defined as (p1, p2)
    and (p3, p4).
    
    Args:
        p1: (float, float)
            Point 1 as (x, y) coordinates.
        
        p2: (float, float)
            Point 2 as (x, y) coordinates.
        
        p3: (float, float)
            Point 3 as (x, y) coordinates.
        
        p4: (float, float)
            Point 4 as (x, y) coordinates.
    
    Returns:
        (float, float) or None
            Coordinates of the intersection point. Returns None if
            there is no intersection.
    """
    
    a_dx = p2[0] - p1[0]
    a_dy = p1[1] - p2[1]
    a_sq = p2[0]*p1[1] - p1[0]*p2[1]
    
    b_dx = p4[0] - p3[0]
    b_dy = p3[1] - p4[1]
    b_sq = p4[0]*p3[1] - p3[0]*p4[1]
    
    d = a_dy * b_dx - a_dx * b_dy
    dx = a_sq * b_dx - a_dx * b_sq
    dy = a_dy * b_sq - a_sq * b_dy
    
    if d == 0:
        return None
    
    return dx/d, dy/d


def intersect_rays(p1, a1, p2, a2):
    """
    Calculates intersection point between two lines defined as a point
    coordinates and angle.
    
    Args:
        p1: (float, float)
            First ray point as (x, y) coordinates.
        
        a1: float
            First ray angle in radians.
        
        p2: (float, float)
            Second ray point as (x, y) coordinates.
        
        a2: float
            Second ray angle in radians.
    
    Returns:
        (float, float) or None
            Coordinates of the intersection point. Returns None if
            there is no intersection.
    """
    
    l1 = (p1[0] + numpy.cos(a1), p1[1] + numpy.sin(a1))
    l2 = (p2[0] + numpy.cos(a2), p2[1] + numpy.sin(a2))
    
    return intersect_lines(p1, l1, p2, l2)


def intersect_line_ray(p1, p2, p3, a):
    """
    Calculates intersection point between two lines defined as a point
    coordinates and angle.
    
    Args:
        p1: (float, float)
            Line point 1 as (x, y) coordinates.
        
        p2: (float, float)
            Line point 2 as (x, y) coordinates.
        
        p3: (float, float)
            Ray point as (x, y) coordinates.
        
        a: float
            Second line angle in radians.
    
    Returns:
        (float, float) or None
            Coordinates of the intersection point. Returns None if
            there is no intersection.
    """
    
    p4 = (p3[0] + numpy.cos(a), p3[1] + numpy.sin(a))
    
    return intersect_lines(p1, p2, p3, p4)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import glyphs
from . glyph import Glyph
from . text import Text, Textbox
from . shapes import Annulus, Arc, Bar, Bow, Ellipse, Line, Polygon, Ray, Rect, Shape, Wedge
from . markers import Marker, Asterisk, Circle, Cross, Diamond, Plus, Triangle, Square, Symbol, MarkerProperty
from . arrows import Arrow, ArcArrow, BowArrow, ConnectorArrow, CurveArrow, LineArrow, PathArrow, RayArrow, ArrowProperty
from . heads import Head, CircleHead, LineHead, NormalHead, OpenHead, SymbolHead, VeeHead, HeadProperty
from . grid import Grid, ParallelGrid, RayGrid, RadialGrid
from . axes import Axis, StraitAxis, RadialAxis
from . gauge import Gauge, StraitGauge, RadialGauge
from . colorbar import ColorBar
from . pather import Pather
from . framer import Framer
from . profile import Profile
from . band import Band
from . labels import LabelBox, Label, TextLabel
from . legend import LegendBox, Legend, MarkerLegend
from . tooltip import Tooltip, TextTooltip

# register properties
from .. import properties
properties.MarkerProperty = MarkerProperty
properties.ArrowProperty = ArrowProperty
properties.HeadProperty = HeadProperty



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Frame, FrameProperty
from . glyph import Glyph


class Label(Glyph):
    """
    Abstract base class for various types of label items glyphs.
    
    The pero.Label classes can be used directly to draw labels or as descriptor
    to create a pero.Label instances from real data and using the 'clone'
    method and a data source.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        x_offset: int, float or callable
            Specifies the x-axis shift from the label anchor.
        
        y_offset: int, float or callable
            Specifies the y-axis shift from the label anchor.
    """
    
    x = NumProperty(UNDEF)
    y = NumProperty(UNDEF)
    
    x_offset = NumProperty(UNDEF)
    y_offset = NumProperty(UNDEF)
    
    
    def get_bbox(self, canvas, source=UNDEF, **overrides):
        """
        Gets glyph bounding box.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
            
            source: any
                Data source to be used for calculating callable properties.
            
            overrides: str:any pairs
                Specific properties to be overwritten.
        
        Returns:
            pero.Frame or None
                Object bounding box.
        """
        
        raise NotImplementedError("The 'get_bbox' method is not implemented for '%s'." % self.__class__.__name__)


class TextLabel(Label):
    """
    Defines a simple text label glyph.
    
    Properties:
        
        text: str, callable, None or UNDEF
            Specifies the text to be drawn.
        
        text properties:
            Includes pero.TextProperties to specify the text properties.
        
        angle properties:
            Includes pero.AngleProperties to specify the text angle.
    """
    
    text = StringProperty(UNDEF)
    font = Include(TextProperties, text_align=TEXT_ALIGN_CENTER, text_base=TEXT_BASE_BOTTOM)
    angle = Include(AngleProperties)
    
    
    def get_bbox(self, canvas, source=UNDEF, **overrides):
        """Gets glyph bounding box."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return None
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        x_offset = self.get_property('x_offset', source, overrides)
        y_offset = self.get_property('y_offset', source, overrides)
        text = self.get_property('text', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # check text
        if not text:
            return None
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        
        # apply offset
        x += x_offset or 0
        y += y_offset or 0
        
        # get bounding box
        return canvas.get_text_bbox(text, x, y, angle)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw label."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        x_offset = self.get_property('x_offset', source, overrides)
        y_offset = self.get_property('y_offset', source, overrides)
        text = self.get_property('text', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # check text
        if not text:
            return
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        
        # check offset
        x_offset = x_offset or 0
        y_offset = y_offset or 0
        
        # draw text
        canvas.draw_text(text, x+x_offset, y+y_offset, angle)


class LabelBox(Glyph):
    """
    Labels box provides a simple tool to draw all given labels at once in
    the order defined by their 'z_index' property. This can be useful in case of
    drawing plot labels as one consistent layer or group.
    
    If the 'clip' frame is provided, all the labels having the anchor
    coordinates outside the frame are ignored and not drawn. In addition, labels
    for which the bounding box falls partially outside the the clipping frame,
    are automatically shifted to ensure their full visibility.
    
    By default the container makes sure the labels do not overlap each other
    using their bounding box. If two labels are overlapping, the one with higher
    'z_index' is finally drawn. To ignore label overlaps the 'overlap' property
    must be set to False.
    
    Properties:
        
        items: (pero.Label,), callable, None or UNDEF
            Specifies a collection of label items to draw.
        
        overlap: bool or callable
            Specifies whether the labels can overlap (True) or should be skipped
            automatically if there is not enough free space (False).
        
        spacing: int, float or callable
            Specifies the minimum free space between adjacent labels.
        
        clip: pero.Frame, callable, None or UNDEF
            Specifies the available drawing frame, which is used to hide
            invisible labels or shift partially visible labels.
        
        padding: int, float, (int,), (float,) callable or UNDEF
            Specifies the inner space as a single value or values for individual
            sides starting from top. This is used in addition to the 'clip' to
            shift partially visible labels.
    """
    
    items = TupleProperty(UNDEF, types=(Label,))
    
    overlap = BoolProperty(False)
    spacing = NumProperty(4)
    
    clip = FrameProperty(UNDEF)
    padding = QuadProperty(5)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw labels."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        
        # get labels as ((label, x, y, x_offset, y_offset),)
        labels = self._get_items(canvas, source, overrides)
        if not labels:
            return
        
        # sort to ensure the most important is drawn on top (as last)
        labels.sort(key=lambda d: d[0].z_index or 0)
        
        # start drawing group
        canvas.group(tag, "labels")
        
        # draw labels
        for label in labels:
            label[0].draw(canvas,
                x = label[1],
                y = label[2],
                x_offset = label[3],
                y_offset = label[4])
        
        # end drawing group
        canvas.ungroup()
    
    
    def _get_items(self, canvas, source, overrides):
        """Gets final list of labels to be drawn."""
        
        # get properties
        items = self.get_property('items', source, overrides)
        overlap = self.get_property('overlap', source, overrides)
        spacing = self.get_property('spacing', source, overrides)
        clip = self.get_property('clip', source, overrides)
        padding = self.get_property('padding', source, overrides)
        
        # check items
        if not items:
            return []
        
        # prepare labels as ((label, x, y, x_offset, y_offset),)
        labels = []
        for label in items:
            x_offset = label.get_property('x_offset') or 0
            y_offset = label.get_property('y_offset') or 0
            labels.append([label, label.x, label.y, x_offset, y_offset])
        
        # allow overlaps and do not clip
        if overlap and not clip:
            return labels
        
        # init padded clip
        padded = clip
        if padding:
            padded = clip.clone()
            padded.shrink(*padding)
        
        # apply clipping and prevent overlaps
        final = []
        area = []
        
        for label in sorted(labels, key=lambda d: d[0].z_index or 0, reverse=True):
            
            # check if outside clip
            if clip and not clip.contains(label[0].x, label[0].y):
                continue
            
            # get bbox
            bbox = label[0].get_bbox(canvas, x=label[1], y=label[2], x_offset=label[3], y_offset=label[4])
            if bbox is None:
                continue
            
            # apply clipping
            if padded:
                
                # calc shift
                x_shift = 0
                y_shift = 0
                
                if bbox.x1 < padded.x1:
                    x_shift = padded.x1 - bbox.x1
                elif bbox.x2 > padded.x2:
                    x_shift = padded.x2 - bbox.x2
                
                if bbox.y1 < padded.y1:
                    y_shift = padded.y1 - bbox.y1
                elif bbox.y2 > padded.y2:
                    y_shift = padded.y2 - bbox.y2
                
                # apply shift
                if x_shift or y_shift:
                    label[3] += x_shift
                    label[4] += y_shift
                    bbox.offset(x_shift, y_shift)
            
            # check overlaps
            if not overlap:
                
                # apply spacing
                if spacing:
                    x, y, width, height = bbox.rect
                    bbox = Frame(x-0.5*spacing, y-0.5*spacing, width+spacing, height+spacing)
                
                # check overlaps
                if any(bbox.overlaps(x) for x in area):
                    continue
            
            # store label
            final.append(label)
            area.append(bbox)
        
        return final



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for logarithmic ticker with base 10."""
    
    
    def test_major_ticks(self):
        """Tests whether major ticks are generated works correctly."""
        
        # init ticker
        ticker = pero.LogTicker()
        
        # test above one
        ticker(start=1.1, end=0.9e3)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1e1, 1e2))
        
        # test one
        ticker(start=1, end=1.1e3)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1, 1e1, 1e2, 1e3))
        
        # test below one
        ticker(start=0.9e-5, end=1.1e-2)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1e-5, 1e-4, 1e-3, 1e-2))
        
        # test cross one
        ticker(start=0.09, end=1.1e3)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (.1, 1, 1e1, 1e2, 1e3))
        
        # test condensed
        ticker(start=1, end=1e10)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1.0, 1e2, 1e4, 1e6, 1e8, 1e10))
        
        # test flipped
        ticker(start=1.1e3, end=0.9)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1e3, 1e2, 1e1, 1))
        
        # test small
        ticker(start=1.1, end=1.2)
        ticks = ticker.major_ticks()
        ticks = tuple(map(lambda x:round(x,2), ticks))
        self.assertEqual(ticks, (1.1, 1.12, 1.14, 1.16, 1.18))
    
    
    def test_minor_ticks(self):
        """Tests whether minor ticks are generated correctly."""
        
        # init ticker
        ticker = pero.LogTicker()
        
        # test step one
        ticker(start=1.1, end=0.9e2)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (2, 3, 4, 5, 6, 7, 8, 9, 10, 20, 30, 40, 50, 60, 70, 80, 90))
        
        # test flipped
        ticker(start=0.9e2, end=1.1)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (90, 80, 70, 60, 50, 40, 30, 20, 10, 9, 8, 7, 6, 5, 4, 3, 2))
        
        # test step grater than one
        ticker(start=1, end=1e14)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (1.0, 10, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9, 1e10, 1e11, 1e12, 1e13, 1e14))
    
    
    def test_beautify(self):
        """Tests whether beautify function works correctly."""
        
        # init ticker
        ticker = pero.LogTicker()
        
        # test above one
        beautified = ticker.beautify(1.1, 0.9e3)
        self.assertEqual(beautified, (1, 1000))
        
        # test one
        beautified = ticker.beautify(1, 0.9e3)
        self.assertEqual(beautified, (1, 1000))
        
        # test below one
        beautified = ticker.beautify(0.9e-5, 1.1e-2)
        self.assertEqual(beautified, (1e-6, 1e-1))
        
        # test cross one
        beautified = ticker.beautify(0.9, 1.1e3)
        self.assertEqual(beautified, (0.1, 10000))
        
        # test flipped
        beautified = ticker.beautify(0.9e3, 0.9)
        self.assertEqual(beautified, (1000, 0.1))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . loader import QWidget, QHBoxLayout
from ... drawing import Graphics
from .. view import Control
from . view import QtView


class QtViewer(QWidget):
    """Simple graphics viewer application."""
    
    
    def __init__(self):
        """Initializes a new instance of QtViewer."""
        
        # init base
        super().__init__()
        self.setWindowTitle("Pero")
        self.resize(750, 500)
        
        # init layout
        self._sizer = QHBoxLayout()
        self._sizer.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._sizer)
        
        # init view
        self._view = None
    
    
    def set_size(self, size):
        """
        Sets app window size.
        
        Args:
            size: (float, float)
                App window size as (width, height).
        """
        
        # set size
        self.resize(size[0], size[1])
        
        # center on screen
        geo = self.frameGeometry()
        center = self.screen().availableGeometry().center()
        geo.moveCenter(center)
        self.move(geo.topLeft())
    
    
    def set_title(self, title):
        """
        Sets app window title.
        
        Args:
            title: str
                App window title.
        """
        
        self.setWindowTitle(title or "")
    
    
    def set_content(self, content):
        """
        Sets content to draw.
        
        Args:
            content: pero.QtView, pero.Control or pero.Graphics
                A content to set.
        """
        
        # clean sizer
        if self._view is not None:
            self._sizer.removeWidget(self._view)
        
        # init view
        if isinstance(content, QtView):
            self._view = content
        
        elif isinstance(content, Control):
            self._view = QtView(self)
            self._view.set_control(content)
        
        elif isinstance(content, Graphics):
            self._view = QtView(self)
            self._view.set_control(Control(graphics=content))
        
        else:
            message = "Unknown content type! -> %s" % type(content)
            raise TypeError(message)
        
        # add to sizer
        self._sizer.addWidget(self._view, 1)
        
        # set focus
        self._view.setFocus()
    
    
    def refresh(self):
        """Redraws view."""
        
        if self._view is not None:
            self._view.refresh()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from . loader import QPoint, QLineF, QColor, QPen, QBrush, QPainterPath, QFontMetrics
from ... properties import *
from ... drawing import Canvas, Path, Matrix, ClipState
from . enums import *


class QtCanvas(Canvas):
    """Wrapper for PyQt drawing context."""
    
    
    def __init__(self, dc, **overrides):
        """
        Initializes a new instance of WXCanvas.
        
        Args:
            dc: QPainter
                PyQt drawing context.
            
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # init buffers
        self._dc = dc
        
        self._pen = self._dc.pen()
        self._brush = self._dc.brush()
        self._default_font = self._dc.font()
        self._font_metrics = QFontMetrics(self._default_font)
        self._for_color = None
        self._bgr_color = None
        
        # set font factor
        self._font_factor = 1.
        if sys.platform == 'win32':
            self._font_factor = 0.75
        
        # init size
        rect = self._dc.viewport()
        
        if 'width' not in overrides:
            overrides['width'] = rect.width()
        
        if 'height' not in overrides:
            overrides['height'] = rect.height()
        
        # init base
        super().__init__(**overrides)
        
        # init canvas
        self._update_pen()
        self._update_brush()
        self._update_text()
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def get_line_size(self, text):
        """
        Gets width and height of a single text line using current text settings.
        
        Args:
            text: str
                Text for which the size should be calculated.
        
        Returns:
            (float, float)
                Line width and height.
        """
        
        # check text
        if not text:
            return 0, 0
        
        # get size
        size = self._font_metrics.boundingRect(text)
        
        return size.width(), size.height()
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        self.draw_ellipse(x, y, 2*radius, 2*radius)
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Draws an ellipse centered around given coordinates and fitting into the
        width and height.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # draw
        self._dc.drawEllipse(int(x-0.5*width), int(y-0.5*height), int(width), int(height))
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # apply scaling and offset
        x1 = self._scale * (x1 + self._offset[0])
        y1 = self._scale * (y1 + self._offset[1])
        x2 = self._scale * (x2 + self._offset[0])
        y2 = self._scale * (y2 + self._offset[1])
        
        # draw
        self._dc.drawLine(int(x1), int(y1), int(x2), int(y2))
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # get lines
        lines = []
        for i in range(0, len(points)-1):
            lines.append(QLineF(points[i][0], points[i][1], points[i+1][0], points[i+1][1]))
        
        # draw
        self._dc.drawLines(*lines)
    
    
    def draw_path(self, path):
        """
        Draws the path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # make qt path
        qt_path = self._make_native_path(path)
        
        # draw path
        self._dc.drawPath(qt_path)
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 3:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # draw
        self._dc.drawPolygon(*(QPoint(int(p[0]), int(p[1])) for p in points))
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corner starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # get radius
        if isinstance(radius, (int, float)) and radius != 0:
            radius = (radius, radius, radius, radius)
        
        # draw as path if different corners
        if radius and not all(r == radius[0] for r in radius):
            path = Path()
            path.rect(x, y, width, height, radius)
            self.draw_path(path)
            return
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width + .5
        height = self._scale * height + .5
        
        # no round corners
        if not radius:
            self._dc.drawRect(int(x), int(y), int(width), int(height))
        
        # same radius for all corners
        else:
            radius = int(self._scale * radius[0])
            self._dc.drawRoundedRect(int(x), int(y), int(width), int(height), int(radius), int(radius))
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # set colors
        ori_pen = self._dc.pen()
        ori_brush = self._dc.brush()
        
        if self._for_color is not None:
            self._dc.setPen(self._for_color)
        
        if self._bgr_color is not None:
            self._dc.setBrush(self._bgr_color)
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        
        # apply angle transformation
        if angle:
            
            x = self._scale * (x + self._offset[0])
            y = self._scale * (y + self._offset[1])
            
            self._dc.save()
            self._dc.translate(x, y)
            self._dc.rotate(numpy.rad2deg(angle))
            
            x = 0
            y = 0
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw lines
        for i, line in enumerate(lines):
            
            # get line size
            line_width, line_height = self.get_line_size(line)
            line_width /= self._scale
            line_height /= self._scale
            ascent = self._font_metrics.ascent()
            
            # init offset
            x_offset = 0
            y_offset = ascent / self._scale
            
            # adjust alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x_offset -= 0.5*line_width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x_offset -= line_width
            
            # adjust baseline
            if self.text_base == TEXT_BASE_MIDDLE:
                y_offset -= 0.5*full_height
            
            elif self.text_base == TEXT_BASE_BOTTOM:
                y_offset -= full_height
            
            # add line offset
            y_offset += i * line_height * (1 + self.text_spacing)
            
            # apply scaling and offset
            if angle:
                line_x = self._scale * x_offset
                line_y = self._scale * y_offset
            else:
                line_x = self._scale * (x + x_offset + self._offset[0])
                line_y = self._scale * (y + y_offset + self._offset[1])
            
            # draw background
            if self._bgr_color is not None:
                
                bgr_x = line_x
                bgr_y = line_y - ascent
                bgr_width = line_width * self._scale
                bgr_height = line_height * self._scale
                
                self._dc.setPen(Qt.PenStyle.NoPen)
                self._dc.drawRect(int(bgr_x), int(bgr_y), int(bgr_width), int(bgr_height))
                
                if self._for_color is not None:
                    self._dc.setPen(self._for_color)
            
            # draw text
            self._dc.drawText(int(line_x), int(line_y), line)
        
        # revert angle transformation
        if angle:
            self._dc.restore()
        
        # revert colors
        self._dc.setPen(ori_pen)
        self._dc.setBrush(ori_brush)
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # save current canvas state
        self._dc.save()
        
        # add path
        qt_path = self._make_native_path(path)
        
        # set as clipping
        self._dc.setClipPath(qt_path, Qt.ClipOperation.IntersectClip)
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # restore canvas state
        self._dc.restore()
        
        # set pen, brush and text
        self._update_pen()
        self._update_brush()
        self._update_text()
    
    
    def _make_native_path(self, path):
        """Converts given path to native path."""
        
        # init path
        qt_path = QPainterPath()
        qt_path.setFillRule(QT_FILL_RULE[path.fill_rule])
        
        # apply commands
        for command in path.commands():
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                qt_path.closeSubpath()
            
            # move to
            elif key == PATH_MOVE:
                qt_path.moveTo(*values)
            
            # line to
            elif key == PATH_LINE:
                qt_path.lineTo(*values)
            
            # curve to
            elif key == PATH_CURVE:
                qt_path.cubicTo(*values)
        
        return qt_path
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._pen.setColor(QColor(*color.rgba))
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._pen.setWidth(int(line_width * self.line_scale))
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._pen.setCapStyle(QT_LINE_CAP[line_cap])
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._pen.setJoinStyle(QT_LINE_JOIN[line_join])
        
        # update style/dash
        if prop_name is None or prop_name in ('line_dash', 'line_style'):
            line_style = self.line_style
            line_dash = self.line_dash if self.line_dash else []
            
            if line_style == LINE_STYLE_SOLID:
                self._pen.setStyle(QT_LINE_STYLE[line_style])
                line_dash = []
            elif line_style == LINE_STYLE_CUSTOM:
                self._pen.setStyle(QT_LINE_STYLE[line_style])
            elif line_style is not UNDEF:
                self._pen.setStyle(QT_LINE_STYLE[LINE_STYLE_CUSTOM])
                line_dash = QT_LINE_STYLE[line_style]
            
            self._pen.setDashPattern(line_dash)
        
        # set pen
        if self._pen.width() == 0:
            self._dc.setPen(Qt.PenStyle.NoPen)
        else:
            self._dc.setPen(self._pen)
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('fill_color', 'fill_alpha'):
            color = ColorProperties.get_color(self, "fill_")
            if color is not UNDEF:
                self._brush.setColor(QColor(*color.rgba))
        
        # update style
        if prop_name is None or prop_name == 'fill_style':
            fill_style = self.fill_style
            if fill_style is not UNDEF:
                self._brush.setStyle(QT_FILL_STYLE[fill_style])
        
        # set back
        self._dc.setBrush(self._brush)
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # get current font
        font = self._dc.font()
        
        # update font name
        if prop_name is None or prop_name == 'font_name':
            font_name = self.font_name
            if font_name is None:
                font.setFamily(self._default_font.family())
            elif font_name is not UNDEF:
                font.setFamily(font_name)
        
        # update font family
        if prop_name is None or prop_name == 'font_family':
            font_family = self.font_family
            if font_family is None:
                font.setFamily(self._default_font.family())
            elif font_family is not UNDEF:
                font.setFamily(QT_FONT_FAMILY[font_family])
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            if font_size is None:
                font.setPointSize(self._default_font.pointSize() * self.font_scale * self._font_factor)
            elif font_size is not UNDEF:
                font.setPointSize(int(font_size * self.font_scale * self._font_factor))
        
        # update font style
        if prop_name is None or prop_name == 'font_style':
            font_style = self.font_style
            if font_style is None:
                font.setStyle(self._default_font.style())
            elif font_style is not UNDEF:
                font.setStyle(QT_FONT_STYLE[font_style])
        
        # update font weight
        if prop_name is None or prop_name == 'font_weight':
            font_weight = self.font_weight
            if font_weight is None:
                font.setWeight(self._default_font.weight())
            elif font_weight is not UNDEF:
                font.setWeight(QT_FONT_WEIGHT[font_weight])
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            self._for_color = QPen(QColor(*color.rgba)) if color.alpha else None
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            color = ColorProperties.get_color(self, "text_bgr_")
            self._bgr_color = QBrush(QColor(*color.rgba)) if color.alpha else None
        
        # set font
        self._dc.setFont(font)
        self._font_metrics = QFontMetrics(font)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import os.path
from .. enums import *


def show(graphics, title=None, width=None, height=None, backend=None, **options):
    """
    Shows given graphics in available viewer app. Currently, this is only
    available if wxPython, PyQt5, PyQt6, PySide2 or PySide2 is installed or
    within Pythonista app on iOS. This method makes sure appropriate backend
    canvas is created and provided to graphics 'draw' method.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Viewer width in device units.
        
        height: float or None
            Viewer height in device units.
        
        backend: pero.BACKEND
            Specific backend to be used. The value must be an item from the
            pero.BACKEND enum.
        
        options: str:any pairs
            Additional parameters for specific backend.
    """
    
    # get backends
    backends = VIEWER_PRIORITY if backend is None else [backend]
    
    # get backend
    backend = None
    for module in backends:
        
        # try to import backend
        try:
            if module == BACKEND_PYTHONISTA:
                from . import pythonista as backend
            
            elif module == BACKEND_QT:
                from . import qt as backend
            
            elif module == BACKEND_WX:
                from . import wx as backend
            
            elif module == BACKEND_PY5:
                from . import py5 as backend
            
            break
        
        # ignore missing library
        except ImportError:
            backend = None
            pass
    
    # no viewer available
    if backend is None:
        message = "No viewer available or missing library (e.g. wxPython, PyQt5, PyQt6, PySide2, PySide6 or Pythonista)!"
        raise ImportError(message)
    
    # show viewer
    backend.show(graphics, title, width, height, **options)


def export(graphics, path, width=None, height=None, backend=None, **options):
    """
    Draws given graphics into specified image file using the format determined
    automatically from the file extension. This method makes sure appropriate
    backend canvas is created and provided to graphics 'draw' method.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        backend: pero.BACKEND
            Specific backend to be used. The value must be an item from the
            pero.BACKEND enum.
        
        options: str:any pairs
            Additional parameters for specific backend.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # get backends
    backends = EXPORT_PRIORITY if backend is None else [backend]
    
    # import backend
    backend = None
    for module in backends:
        
        # check if format is recognized by backend
        if extension not in EXPORT_FORMATS[module]:
            continue
        
        # try to import backend
        try:
            if module == BACKEND_CAIRO:
                from . import cairo as backend
            
            elif module == BACKEND_JSON:
                from . import json as backend
            
            elif module == BACKEND_MUPDF:
                from . import mupdf as backend
            
            elif module == BACKEND_PYTHONISTA:
                from . import pythonista as backend
            
            elif module == BACKEND_QT:
                from . import qt as backend
            
            elif module == BACKEND_SVG:
                from . import svg as backend
            
            elif module == BACKEND_WX:
                from . import wx as backend
            
            elif module == BACKEND_PY5:
                from . import py5 as backend
            
            break
        
        # ignore missing library
        except ImportError:
            backend = None
            pass
    
    # unsupported format
    if backend is None:
        message = "Unsupported image format or missing library (e.g. wxPython, PyCairo, PyQt5, PyQt6, PySide2, PySide6)! -> %s" % extension
        raise ImportError(message)
    
    # export image
    backend.export(graphics, path, width, height, **options)


def debug(graphics, canvas='show', title="", width=None, height=None, backend=None, **options):
    """
    Renders given graphics using simple viewer or file format. This method makes
    sure appropriate backend canvas is created and provided to graphics 'draw'
    method. In case of the image export the file is stored in current working
    directory using 'test' and appropriate format as the name.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        canvas: str
            Specifies drawing mechanism to be used. This can be either 'show',
            'wx', 'qt' or 'pythonista' to display graphics within a simple
            viewer app, or any of the supported image formats (e.g. 'png',
            'svg') to draw to 'test.?' file.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Image or viewer width in device units.
        
        height: float or None
            Image or viewer height in device units.
        
        backend: pero.BACKEND
            Specific backend to be used. The value must be an item from the
            pero.BACKEND enum.
        
        options: key:value pairs
            Additional parameters for specific backend.
    """
    
    # render graphics in available viewer
    if canvas == 'show':
        show(graphics, title, width, height, backend, **options)
    
    # render graphics into qt viewer
    elif canvas == BACKEND_QT:
        from . import qt
        qt.show(graphics, title, width, height, **options)
    
    # render graphics into wx viewer
    elif canvas == BACKEND_WX:
        from . import wx
        wx.show(graphics, title, width, height, **options)
    
    # render graphics into Pythonista console
    elif canvas == BACKEND_PYTHONISTA:
        from . import pythonista
        pythonista.show(graphics, title, width, height, **options)
    
    # render graphics into Py5 sketch
    elif canvas == BACKEND_PY5:
        from . import py5
        py5.show(graphics, title, width, height, **options)
    
    # render graphics as image file
    else:
        filename = "test.%s" % canvas
        export(graphics, filename, width, height, backend, **options)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from . undefined import UNDEF
from . prop import Property


class EnumProperty(Property):
    """Defines a generic property allowing predefined set of values only."""
    
    
    def __init__(self, default=UNDEF, enum=(), **kwargs):
        """
        Initializes a new instance of EnumProperty.
        
        Args:
            enum: pero.Enum or (any,)
                Allowed values.
        """
        
        self._enum = set(enum)
        
        kwargs['default'] = default
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check allowed values
        if value in self._enum:
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # cannot convert value
        enums = "|".join(str(x) for x in self._enum)
        message = "Value of '%s' property must be in (%s)! -> %s" % (self.name, enums, value)
        raise TypeError(message)
    
    
    def clone(self, **kwargs):
        """Creates a clone of current property."""
        
        return super().clone(enum=self._enum, **kwargs)


class RangeProperty(Property):
    """Defines a generic property allowing values within specific range only."""
    
    
    def __init__(self, default=UNDEF, minimum=None, maximum=None, minimum_incl=True, maximum_incl=True, **kwargs):
        """
        Initializes a new instance of RangeProperty.
        
        Args:
            minimum: int, float or None
                Minimum allowed value.
            
            maximum: int, float or None
                Maximum allowed value.
            
            minimum_incl: bool
                Specifies whether the minimum value should be included (True)
                or excluded (False) from allowed range.
            
            maximum_incl: bool
                Specifies whether the maximum value should be included (True)
                or excluded (False) from allowed range.
        """
        
        self._minimum = minimum
        self._maximum = maximum
        self._minimum_incl = minimum_incl
        self._maximum_incl = maximum_incl
        
        kwargs['default'] = default
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # check minimum
        if self._minimum is not None:
            
            if not self._minimum_incl and value <= self._minimum:
                message = "Value of '%s' property must be greater than %s! -> %s" % (self.name, self._minimum, value)
                raise ValueError(message)
            
            elif value < self._minimum:
                message = "Value of '%s' property must be greater or equal to %s! -> %s" % (self.name, self._minimum, value)
                raise ValueError(message)
        
        # check maximum
        if self._maximum is not None:
            
            if not self._maximum_incl and value >= self._maximum:
                message = "Value of '%s' property must be smaller than %s! -> %s" % (self.name, self._maximum, value)
                raise ValueError(message)
            
            elif value > self._maximum:
                message = "Value of '%s' property must be smaller or equal to %s! -> %s" % (self.name, self._maximum, value)
                raise ValueError(message)
        
        return value
    
    
    def clone(self, **kwargs):
        """Creates a clone of current property."""
        
        return super().clone(
            minimum = self._minimum,
            maximum = self._maximum,
            minimum_incl = self._minimum_incl,
            maximum_incl = self._maximum_incl,
            **kwargs)


class FuncProperty(Property):
    """Defines a callable property."""
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # allow functions
        if callable(value):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # cannot convert value
        message = "Value of '%s' property must be callable! -> %s" % (self.name, type(value))
        raise TypeError(message)


class BoolProperty(Property):
    """Defines a boolean property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of BoolProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (bool,)
        
        super().__init__(**kwargs)


class IntProperty(Property):
    """Defines an integer property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of IntProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (int,)
        
        super().__init__(**kwargs)


class FloatProperty(Property):
    """Defines a float property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of FloatProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (float,)
        
        super().__init__(**kwargs)


class NumProperty(Property):
    """Defines a numeric property for float and int."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of NumProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (int, float)
        
        super().__init__(**kwargs)


class StringProperty(Property):
    """Defines a string property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of StringProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (str,)
        
        super().__init__(**kwargs)


class IterProperty(Property):
    """Defines a collection property with specified types of inner elements."""
    
    
    def __init__(self, default=UNDEF, intypes=(), **kwargs):
        """
        Initializes a new instance of IterProperty.
        
        Args:
            intypes: (Type,)
                Allowed types for collection elements. If empty, specific inner
                type is not required for inner elements.
        """
        
        # set allowed inner types
        self._intypes = intypes
        
        if isinstance(intypes, list):
            self._intypes = tuple(intypes)
        
        elif not isinstance(intypes, tuple):
            self._intypes = (intypes,)
        
        # init property
        kwargs['default'] = default
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # skip inner type checking
        if not self._intypes:
            return value
        
        # check inner types
        for elm in value:
            if not isinstance(elm, self._intypes):
                intypes = "|".join(x.__name__ for x in self._intypes)
                message = "All elements of the '%s' property must be of type (%s)! -> %s" % (self.name, intypes, type(elm))
                raise TypeError(message)
        
        return value
    
    
    def clone(self, **kwargs):
        """Creates a clone of current property."""
        
        return super().clone(intypes=self._intypes, **kwargs)


class SequenceProperty(IterProperty):
    """Defines a sequence property allowing list, tuple and numpy.ndarray."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of ListProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (list, tuple, numpy.ndarray)
        
        super().__init__(**kwargs)


class ListProperty(IterProperty):
    """Defines a list property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of ListProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (list,)
        
        super().__init__(**kwargs)


class TupleProperty(IterProperty):
    """Defines a tuple property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of TupleProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (tuple,)
        
        super().__init__(**kwargs)


class SetProperty(IterProperty):
    """Defines a set property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of SetProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (set,)
        
        super().__init__(**kwargs)


class DictProperty(Property):
    """Defines a dictionary property."""
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of DictProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (dict,)
        
        super().__init__(**kwargs)


class QuadProperty(Property):
    """
    Defines a property used for quad-sided connected values like margins,
    paddings etc. The value must be provided as a single number or as a list or
    tuple of numbers defining individual values for top, right, bottom and left
    side. If only a single value is provided, it is used for all the sides. If
    two values are provided, the first value is set to top and bottom and the
    second value is set to left and right.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of QuadProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (int, float, tuple, list)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # single value
        if isinstance(value, (int, float)):
            return value, value, value, value
        
        # check type
        if isinstance(value, (list,tuple)) and all(isinstance(x, (int, float)) for x in value):
            
            if len(value) == 1:
                return value[0], value[0], value[0], value[0]
            
            elif len(value) == 2:
                return value[0], value[1], value[0], value[1]
            
            elif len(value) == 4:
                return tuple(value)
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # wrong value
        message = "Value of '%s' property must be a number or list or tuple of 2 or 4 numbers! -> %s" % (self.name, value)
        raise TypeError(message)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from .. drawing import Matrix, Path
from . glyph import Glyph


class Marker(Glyph):
    """
    Abstract base class for various types of markers, typically used to draw
    individual points of plotted series etc. They are simple glyphs defined by
    size and drawn at specified position.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the center.
        
        y: int, float or callable
            Specifies the y-coordinate of the center.
        
        size: int, float or callable
            Specifies the size.
        
        line properties:
            Includes pero.LineProperties to specify the marker outline.
        
        fill properties:
            Includes pero.FillProperties to specify the marker fill.
    """
    
    TYPE = None
    
    x = NumProperty(0)
    y = NumProperty(0)
    size = NumProperty(8)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    @staticmethod
    def create(symbol, **overrides):
        """
        Initializes a new instance of specified marker.
        
        Args:
            symbol: str or pero.Path
                Marker symbol.
        
        Returns:
            pero.Marker
                Initialized marker.
        """
        
        # convert from symbol
        if symbol == MARKER_ASTERISK:
            return Asterisk(**overrides)
        
        elif symbol == MARKER_CIRCLE:
            return Circle(**overrides)
        
        elif symbol == MARKER_CROSS:
            return Cross(**overrides)
        
        elif symbol == MARKER_PLUS:
            return Plus(**overrides)
        
        elif symbol == MARKER_TRIANGLE:
            return Triangle(**overrides)
        
        elif symbol == MARKER_SQUARE:
            return Square(**overrides)
        
        elif symbol == MARKER_DIAMOND:
            return Diamond(**overrides)
        
        elif symbol == MARKER_PENTAGON:
            marker = Symbol(path=Path.make_ngon(5), **overrides)
            marker.TYPE = MARKER_PENTAGON
            return marker
        
        elif symbol == MARKER_HEXAGON:
            marker = Symbol(path=Path.make_ngon(6), **overrides)
            marker.TYPE = MARKER_HEXAGON
            return marker
        
        elif isinstance(symbol, Path):
            return Symbol(path=symbol, **overrides)
        
        raise ValueError("Unknown marker symbol! -> '%s'" % symbol)


class Asterisk(Marker):
    """Defines an asterisk marker."""
    
    TYPE = MARKER_ASTERISK
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # get radii
        radius1 = 0.5*size
        radius2 = radius1*0.5*numpy.sqrt(2)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_line(x-radius1, y, x+radius1, y)
        canvas.draw_line(x, y-radius1, x, y+radius1)
        canvas.draw_line(x-radius2, y-radius2, x+radius2, y+radius2)
        canvas.draw_line(x-radius2, y+radius2, x+radius2, y-radius2)


class Circle(Marker):
    """Defines a circular marker."""
    
    TYPE = MARKER_CIRCLE
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_circle(x, y, 0.5*size)


class Cross(Marker):
    """Defines a cross marker."""
    
    TYPE = MARKER_CROSS
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # get radius
        radius = 0.5*size
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_line(x-radius, y-radius, x+radius, y+radius)
        canvas.draw_line(x+radius, y-radius, x-radius, y+radius)


class Diamond(Marker):
    """Defines a diamond marker."""
    
    TYPE = MARKER_DIAMOND
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # get radius
        radius = 0.5*size
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_polygon((
            (x, y-radius),
            (x+radius, y),
            (x, y+radius),
            (x-radius, y)))


class Plus(Marker):
    """Defines a plus marker."""
    
    TYPE = MARKER_PLUS
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # get radius
        radius = 0.5*size
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_line(x-radius, y, x+radius, y)
        canvas.draw_line(x, y-radius, x, y+radius)


class Triangle(Marker):
    """Defines a triangle marker."""
    
    TYPE = MARKER_TRIANGLE
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # get factor
        f = (size/2.)/(numpy.sqrt(3)/2.)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_polygon((
            (x, y-f),
            (x+0.5*size, y+f*0.5),
            (x-0.5*size, y+f*0.5)))


class Square(Marker):
    """Defines a square marker."""
    
    TYPE = MARKER_SQUARE
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_rect(x-0.5*size, y-0.5*size, size, size)


class Symbol(Marker):
    """
    Defines a custom path marker. The path must be a 'symbol-path' i.e. centered
    at 0,0 and scaled to fit into 1x1 square. This can be created from any path
    by calling the 'symbol' method.
    
    Properties:
        
        path: pero.Path, callable, None or UNDEF
            Specifies the 'symbol-path' to be used as marker.
    """
    
    path = Property(UNDEF, types=(Path,), nullable=True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw marker."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        size = self.get_property('size', source, overrides)
        path = self.get_property('path', source, overrides)
        
        # check data
        if not path:
            return
        
        # scale path
        matrix = Matrix().scale(size, size).translate(x, y)
        path = path.transformed(matrix)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class MarkerProperty(Property):
    """
    Defines a marker property, which simplifies a marker definition by
    converting specific symbols or path into an instance of corresponding marker
    glyph. Available symbols are defined by the pero.MARKER enum.
    
    The path must be a 'symbol-path' i.e. centered at 0,0 and scaled to fit into
    1x1 square. This can be created from any path by calling the 'symbol' method.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of MarkerProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Marker, str, Path)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check type
        if isinstance(value, Marker):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert to marker
        return Marker.create(value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for custom function formatter."""
    
    
    def test_exact(self):
        """Tests whether formatter works for exact indices."""
        
        labels = ('zero', 'one', 'two', 'three', 'four')
        formatter = pero.IndexFormatter(labels=labels)
        
        self.assertEqual(formatter.format(0), 'zero')
        self.assertEqual(formatter.format(1), 'one')
        self.assertEqual(formatter.format(2), 'two')
        self.assertEqual(formatter.format(3), 'three')
        self.assertEqual(formatter.format(4), 'four')
    
    
    def test_rounding(self):
        """Tests whether formatter works for floats."""
        
        labels = ('zero', 'one', 'two', 'three', 'four')
        formatter = pero.IndexFormatter(labels=labels, default=pero.UNDEF)
        
        self.assertEqual(formatter.format(2.2), 'two')
        self.assertEqual(formatter.format(2.7), 'three')
        self.assertEqual(formatter.format(4.6), pero.UNDEF)
    
    
    def test_outside(self):
        """Tests whether formatter works for outside indices."""
        
        labels = ('zero', 'one', 'two', 'three', 'four')
        formatter = pero.IndexFormatter(labels=labels, default=pero.UNDEF)
        
        self.assertEqual(formatter.format(-1), pero.UNDEF)
        self.assertEqual(formatter.format(5), pero.UNDEF)
    
    
    def test_default(self):
        """Tests whether default value works correctly."""
        
        labels = ('zero', 'one', 'two', 'three', 'four')
        formatter = pero.IndexFormatter(labels=labels, default='joker')
        
        self.assertEqual(formatter.format(-1), 'joker')
        self.assertEqual(formatter.format(5), 'joker')


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)
